# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras python-based idempotent saving functions (experimental)."""
import importlib
import json
import os
import types

import tensorflow.compat.v2 as tf

from keras.saving.saved_model import json_utils
from keras.utils import generic_utils

# isort: off
from tensorflow.python.util import tf_export

_CONFIG_FILE = "config.keras"

# A temporary flag to enable the new idempotent saving framework.
_ENABLED = False


def load(dirpath):
    """Load a saved python model."""
    file_path = os.path.join(dirpath, _CONFIG_FILE)
    with tf.io.gfile.GFile(file_path, "r") as f:
        config_json = f.read()
    config_dict = json_utils.decode(config_json)
    return deserialize_keras_object(config_dict)


def save(model, dirpath):
    """Save a saved python model."""
    if not tf.io.gfile.exists(dirpath):
        tf.io.gfile.mkdir(dirpath)
    file_path = os.path.join(dirpath, _CONFIG_FILE)

    # TODO(rchao): Save the model's metadata (e.g. Keras version) in a separate
    # file in the archive.
    # TODO(rchao): Save the model's state (e.g. layer weights/vocab) in a
    # separate set of files in the archive.
    # TODO(rchao): Write the config into a file in an archive. In this prototype
    # we're temporarily settled on a standalone json file.
    serialized_model_dict = serialize_keras_object(model)
    config_json = json.dumps(serialized_model_dict, cls=json_utils.Encoder)
    with tf.io.gfile.GFile(file_path, "w") as f:
        f.write(config_json)


# TODO(rchao): Replace the current Keras' `deserialize_keras_object` with this
# (as well as the reciprocal function).
def deserialize_keras_object(config_dict):
    """Retrieve the object by deserializing the config dict.

    The config dict is a python dictionary that consists of a set of key-value
    pairs, and represents a Keras object, such as an `Optimizer`, `Layer`,
    `Metrics`, etc. The saving and loading library uses the following keys to
    record information of a Keras object:

    - `class_name`: String. For classes that have an exported Keras namespace,
      this is the full path that starts with "keras", such as
      "keras.optimizers.Adam". For classes that do not have an exported Keras
      namespace, this is the name of the class, as exactly defined in the source
      code, such as "LossesContainer".
    - `config`: Dict. Library-defined or user-defined key-value pairs that store
      the configuration of the object, as obtained by `object.get_config()`.
    - `module`: String. The path of the python module, such as
      "keras.engine.compile_utils". Built-in Keras classes
      expect to have prefix `keras`. For classes that have an exported Keras
      namespace, this is `None` since the class can be fully identified by the
      full Keras path.
    - `registered_name`: String. The key the class is registered under via
      `keras.utils.register_keras_serializable(package, name)` API. The key has
      the format of '{package}>{name}', where `package` and `name` are the
      arguments passed to `register_keras_serializable()`. If `name` is not
      provided, it defaults to the class name. If `registered_name` successfully
      resolves to a class (that was registered), `class_name` and `config`
      values in the dict will not be used. `registered_name` is only used for
      non-built-in classes.

    For example, the following dictionary represents the built-in Adam optimizer
    with the relevant config. Note that for built-in (exported symbols that have
    an exported Keras namespace) classes, the library tracks the class by the
    the import location of the built-in object in the Keras namespace, e.g.
    `"keras.optimizers.Adam"`, and this information is stored in `class_name`:

    ```
    dict_structure = {
        "class_name": "keras.optimizers.Adam",
        "config": {
            "amsgrad": false,
            "beta_1": 0.8999999761581421,
            "beta_2": 0.9990000128746033,
            "decay": 0.0,
            "epsilon": 1e-07,
            "learning_rate": 0.0010000000474974513,
            "name": "Adam"
        },
        "module": null,
        "registered_name": "Adam"
    }
    # Returns an `Adam` instance identical to the original one.
    deserialize_keras_object(dict_structure)
    ```

    If the class does not have an exported Keras namespace, the library tracks
    it by its `module` and `class_name`. For example:

    ```
    dict_structure = {
      "class_name": "LossesContainer",
      "config": {
          "losses": [...],
          "total_loss_mean": {...},
      },
      "module": "keras.engine.compile_utils",
      "registered_name": "LossesContainer"
    }

    # Returns a `LossesContainer` instance identical to the original one.
    deserialize_keras_object(dict_structure)
    ```

    And the following dictionary represents a user-customized `MeanSquaredError`
    loss:

    ```
    @keras.utils.generic_utils.register_keras_serializable(package='my_package')
    class ModifiedMeanSquaredError(keras.losses.MeanSquaredError):
      ...

    dict_structure = {
        "class_name": "ModifiedMeanSquaredError",
        "config": {
            "fn": "mean_squared_error",
            "name": "mean_squared_error",
            "reduction": "auto"
        },
        "registered_name": "my_package>ModifiedMeanSquaredError"
    }
    # Gives `ModifiedMeanSquaredError` object
    deserialize_keras_object(dict_structure)
    ```

    Args:
      config_dict: the python dict structure to deserialize the Keras object
        from.

    Returns:
      The Keras object that is deserialized from `config_dict`.

    """
    # TODO(rchao): Design a 'version' key for `config_dict` for defining
    # versions for classes.
    class_name = config_dict["class_name"]
    config = config_dict["config"]
    module = config_dict["module"]
    registered_name = config_dict["registered_name"]

    # Strings and functions will have `builtins` as its module.
    if module == "builtins":
        if class_name == "str":
            if not isinstance(config, str):
                raise TypeError(
                    "Config of string is supposed to be a string. "
                    f"Received: {config}."
                )
            return config

        elif class_name == "function":
            custom_function = generic_utils.get_custom_objects_by_name(
                registered_name
            )
            if custom_function is not None:
                # If there is a custom function registered (via
                # `register_keras_serializable` API), that takes precedence.
                return custom_function

            # Otherwise, attempt to import the tracked module, and find the
            # function.
            function_module = config.get("module", None)
            try:
                function_module = importlib.import_module(function_module)
            except ImportError as e:
                raise ImportError(
                    f"The function module {function_module} is not available. "
                    f"The config dictionary provided is {config_dict}."
                ) from e
            return vars(function_module).get(config["function_name"])

        raise TypeError(f"Unrecognized type: {class_name}")

    custom_class = generic_utils.get_custom_objects_by_name(registered_name)
    if custom_class is not None:
        # For others (classes), see if there is a custom class registered (via
        # `register_keras_serializable` API). If so, that takes precedence.
        return custom_class.from_config(config)
    else:
        # Otherwise, attempt to retrieve the class object given the `module`,
        # and `class_name`.
        if module is None:
            # In the case where `module` is not recorded, the `class_name`
            # represents the full exported Keras namespace (used by
            # `keras_export`) such as "keras.optimizers.Adam".
            cls = tf_export.get_symbol_from_name(class_name)
        else:
            # In the case where `module` is available, the class does not have
            # an Keras namespace (which is the case when the symbol is not
            # exported via `keras_export`). Import the tracked module (that is
            # used for the internal path), find the class, and use its config.
            mod = importlib.import_module(module)
            cls = vars(mod).get(class_name, None)
        if not hasattr(cls, "from_config"):
            raise TypeError(f"Unable to reconstruct an instance of {cls}.")
        return cls.from_config(config)


def serialize_keras_object(obj):
    """Retrieve the config dict by serializing the Keras object.

    `serialize_keras_object()` serializes a Keras object to a python dictionary
    that represents the object, and is a reciprocal function of
    `deserialize_keras_object()`. See `deserialize_keras_object()` for more
    information about the config format.

    Args:
      obj: the Keras object to serialize.

    Returns:
      A python dict that represents the object. The python dict can be
      deserialized via `deserialize_keras_object()`.
    """

    # Note that in the case of the `obj` being a function, the module used will
    # be "builtins", and the `class_name` used will be "function"; in the case
    # of the `obj` being a string, the module used will be "builtins", and the
    # `class_name` used will be "str"
    module = None

    # This gets the `keras.*` exported name, such as "keras.optimizers.Adam".
    class_name = tf_export.get_canonical_name_for_symbol(
        obj.__class__, api_name="keras"
    )
    if class_name is None:
        module = obj.__class__.__module__
        class_name = obj.__class__.__name__
    return {
        "module": module,
        "class_name": class_name,
        "config": _get_object_config(obj),
        "registered_name": _get_object_registered_name(obj),
    }


def _get_object_registered_name(obj):
    if isinstance(obj, types.FunctionType):
        return generic_utils.get_registered_name(obj)
    else:
        return generic_utils.get_registered_name(obj.__class__)


def _get_object_config(obj):
    """Return the object's config depending on string, function, or others."""
    if isinstance(obj, str):
        # Use the content of the string as the config for string.
        return obj
    elif isinstance(obj, types.FunctionType):
        # Keep track of the function's module and name in a dict as the config.
        return {
            "module": obj.__module__,
            "function_name": obj.__name__,
        }
    if not hasattr(obj, "get_config"):
        raise TypeError(f"Unable to recognize the config of {obj}.")
    return obj.get_config()



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras abstract base for depthwise convolutions."""


import tensorflow.compat.v2 as tf

from keras import constraints
from keras import initializers
from keras import regularizers
from keras.engine.input_spec import InputSpec
from keras.layers.convolutional.base_conv import Conv


class DepthwiseConv(Conv):
    """Depthwise convolution.

    Depthwise convolution is a type of convolution in which each input channel
    is convolved with a different kernel (called a depthwise kernel). You can
    understand depthwise convolution as the first step in a depthwise separable
    convolution.

    It is implemented via the following steps:

    - Split the input into individual channels.
    - Convolve each channel with an individual depthwise kernel with
      `depth_multiplier` output channels.
    - Concatenate the convolved outputs along the channels axis.

    Unlike a regular convolution, depthwise convolution does not mix
    information across different input channels.

    The `depth_multiplier` argument determines how many filter are applied to
    one input channel. As such, it controls the amount of output channels that
    are generated per input channel in the depthwise step.

    Args:
      kernel_size: A tuple or list of integers specifying the spatial dimensions
        of the filters. Can be a single integer to specify the same value for
        all spatial dimensions.
      strides: A tuple or list of integers specifying the strides of the
        convolution. Can be a single integer to specify the same value for all
        spatial dimensions. Specifying any `stride` value != 1 is incompatible
        with specifying any `dilation_rate` value != 1.
      padding: One of `"valid"` or `"same"` (case-insensitive). `"valid"` means
        no padding. `"same"` results in padding with zeros evenly to the
        left/right or up/down of the input such that output has the same
        height/width dimension as the input.
      depth_multiplier: The number of depthwise convolution output channels for
        each input channel. The total number of depthwise convolution output
        channels will be equal to `filters_in * depth_multiplier`.
      data_format: A string, one of `channels_last` (default) or
        `channels_first`.  The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape `(batch_size, height,
        width, channels)` while `channels_first` corresponds to inputs with
        shape `(batch_size, channels, height, width)`. It defaults to the
        `image_data_format` value found in your Keras config file at
        `~/.keras/keras.json`. If you never set it, then it will be
        'channels_last'.
      dilation_rate: An integer or tuple/list of 2 integers, specifying the
        dilation rate to use for dilated convolution. Currently, specifying any
        `dilation_rate` value != 1 is incompatible with specifying any `strides`
        value != 1.
      activation: Activation function to use. If you don't specify anything, no
        activation is applied (see `keras.activations`).
      use_bias: Boolean, whether the layer uses a bias vector.
      depthwise_initializer: Initializer for the depthwise kernel matrix (see
        `keras.initializers`). If None, the default initializer
        ('glorot_uniform') will be used.
      bias_initializer: Initializer for the bias vector (see
        `keras.initializers`). If None, the default initializer ('zeros') will
        be used.
      depthwise_regularizer: Regularizer function applied to the depthwise
        kernel matrix (see `keras.regularizers`).
      bias_regularizer: Regularizer function applied to the bias vector (see
        `keras.regularizers`).
      activity_regularizer: Regularizer function applied to the output of the
        layer (its 'activation') (see `keras.regularizers`).
      depthwise_constraint: Constraint function applied to the depthwise kernel
        matrix (see `keras.constraints`).
      bias_constraint: Constraint function applied to the bias vector (see
        `keras.constraints`).

    Input shape:
      4D tensor with shape: `[batch_size, channels, rows, cols]` if
        data_format='channels_first'
      or 4D tensor with shape: `[batch_size, rows, cols, channels]` if
        data_format='channels_last'.

    Output shape:
      4D tensor with shape: `[batch_size, channels * depth_multiplier, new_rows,
        new_cols]` if `data_format='channels_first'`
        or 4D tensor with shape: `[batch_size,
        new_rows, new_cols, channels * depth_multiplier]` if
        `data_format='channels_last'`. `rows` and `cols` values might have
        changed due to padding.

    Returns:
      A tensor of rank 4 representing
      `activation(depthwiseconv2d(inputs, kernel) + bias)`.

    Raises:
      ValueError: if `padding` is "causal".
      ValueError: when both `strides` > 1 and `dilation_rate` > 1.
    """

    def __init__(
        self,
        rank,
        kernel_size,
        strides=1,
        padding="valid",
        depth_multiplier=1,
        data_format=None,
        dilation_rate=1,
        activation=None,
        use_bias=True,
        depthwise_initializer="glorot_uniform",
        bias_initializer="zeros",
        depthwise_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        depthwise_constraint=None,
        bias_constraint=None,
        **kwargs,
    ):
        super().__init__(
            rank,
            filters=None,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=activation,
            use_bias=use_bias,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            bias_constraint=bias_constraint,
            **kwargs,
        )
        self.depth_multiplier = depth_multiplier
        self.depthwise_initializer = initializers.get(depthwise_initializer)
        self.depthwise_regularizer = regularizers.get(depthwise_regularizer)
        self.depthwise_constraint = constraints.get(depthwise_constraint)
        self.bias_initializer = initializers.get(bias_initializer)

    def build(self, input_shape):
        if len(input_shape) != self.rank + 2:
            raise ValueError(
                "Inputs to `DepthwiseConv` should have "
                f"rank {self.rank + 2}. "
                f"Received input_shape={input_shape}."
            )
        input_shape = tf.TensorShape(input_shape)
        channel_axis = self._get_channel_axis()
        if input_shape.dims[channel_axis].value is None:
            raise ValueError(
                "The channel dimension of the inputs to `DepthwiseConv` "
                "should be defined. "
                f"The input_shape received is {input_shape}, "
                f"where axis {channel_axis} (0-based) "
                "is the channel dimension, which found to be `None`."
            )
        input_dim = int(input_shape[channel_axis])
        depthwise_kernel_shape = self.kernel_size + (
            input_dim,
            self.depth_multiplier,
        )

        self.depthwise_kernel = self.add_weight(
            shape=depthwise_kernel_shape,
            initializer=self.depthwise_initializer,
            name="depthwise_kernel",
            regularizer=self.depthwise_regularizer,
            constraint=self.depthwise_constraint,
        )

        if self.use_bias:
            self.bias = self.add_weight(
                shape=(input_dim * self.depth_multiplier,),
                initializer=self.bias_initializer,
                name="bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = None
        # Set input spec.
        self.input_spec = InputSpec(
            min_ndim=self.rank + 2, axes={channel_axis: input_dim}
        )
        self.built = True

    def call(self, inputs):
        raise NotImplementedError

    def get_config(self):
        config = super().get_config()
        config.pop("filters")
        config.pop("kernel_initializer")
        config.pop("kernel_regularizer")
        config.pop("kernel_constraint")
        config["depth_multiplier"] = self.depth_multiplier
        config["depthwise_initializer"] = initializers.serialize(
            self.depthwise_initializer
        )
        config["depthwise_regularizer"] = regularizers.serialize(
            self.depthwise_regularizer
        )
        config["depthwise_constraint"] = constraints.serialize(
            self.depthwise_constraint
        )
        return config



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the ActivityRegularization layer."""


from keras import regularizers
from keras.engine.base_layer import Layer

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.ActivityRegularization")
class ActivityRegularization(Layer):
    """Layer that applies an update to the cost function based input activity.

    Args:
      l1: L1 regularization factor (positive float).
      l2: L2 regularization factor (positive float).

    Input shape:
      Arbitrary. Use the keyword argument `input_shape`
      (tuple of integers, does not include the samples axis)
      when using this layer as the first layer in a model.

    Output shape:
      Same shape as input.
    """

    def __init__(self, l1=0.0, l2=0.0, **kwargs):
        super().__init__(
            activity_regularizer=regularizers.L1L2(l1=l1, l2=l2), **kwargs
        )
        self.supports_masking = True
        self.l1 = l1
        self.l2 = l2

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = {"l1": self.l1, "l2": self.l2}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras mixed precision API.

See [the mixed precision guide](
  https://www.tensorflow.org/guide/keras/mixed_precision) to learn how to
use the API.

"""

import sys as _sys

from keras.mixed_precision.loss_scale_optimizer import BaseLossScaleOptimizer as LossScaleOptimizer
from keras.mixed_precision.policy import Policy
from keras.mixed_precision.policy import global_policy
from keras.mixed_precision.policy import set_global_policy


# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Adagrad optimizer implementation."""

import numpy as np
import tensorflow.compat.v2 as tf

from keras import backend_config
from keras.optimizers.optimizer_v2 import optimizer_v2

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.optimizers.Adagrad")
class Adagrad(optimizer_v2.OptimizerV2):
    r"""Optimizer that implements the Adagrad algorithm.

    Adagrad is an optimizer with parameter-specific learning rates,
    which are adapted relative to how frequently a parameter gets
    updated during training. The more updates a parameter receives,
    the smaller the updates.

    Args:
      learning_rate: Initial value for the learning rate:
        either a floating point value,
        or a `tf.keras.optimizers.schedules.LearningRateSchedule` instance.
        Defaults to 0.001.
        Note that `Adagrad` tends to benefit from higher initial learning rate
        values compared to other optimizers.
        To match the exact form in the original paper, use 1.0.
      initial_accumulator_value: Floating point value.
        Starting value for the accumulators (per-parameter momentum values).
        Must be non-negative.
      epsilon: Small floating point value used to maintain numerical stability.
      name: Optional name prefix for the operations created when applying
        gradients.  Defaults to `"Adagrad"`.
      **kwargs: keyword arguments. Allowed arguments are `clipvalue`,
        `clipnorm`, `global_clipnorm`.
        If `clipvalue` (float) is set, the gradient of each weight
        is clipped to be no higher than this value.
        If `clipnorm` (float) is set, the gradient of each weight
        is individually clipped so that its norm is no higher than this value.
        If `global_clipnorm` (float) is set the gradient of all weights is
        clipped so that their global norm is no higher than this value..

    Reference:
      - [Duchi et al., 2011](
        http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf).
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self,
        learning_rate=0.001,
        initial_accumulator_value=0.1,
        epsilon=1e-7,
        name="Adagrad",
        **kwargs
    ):
        if initial_accumulator_value < 0.0:
            raise ValueError(
                "initial_accumulator_value must be non-negative: %s"
                % initial_accumulator_value
            )
        if epsilon is None:
            epsilon = backend_config.epsilon()
        super().__init__(name, **kwargs)
        self._set_hyper("learning_rate", kwargs.get("lr", learning_rate))
        self._set_hyper("decay", self._initial_decay)
        self._initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon or backend_config.epsilon()

    def _create_slots(self, var_list):
        for var in var_list:
            dtype = var.dtype.base_dtype
            init = tf.compat.v1.constant_initializer(
                self._initial_accumulator_value, dtype=dtype
            )
            self.add_slot(var, "accumulator", init)

    def _prepare_local(self, var_device, var_dtype, apply_state):
        super()._prepare_local(var_device, var_dtype, apply_state)
        apply_state[(var_device, var_dtype)].update(
            dict(
                epsilon=tf.convert_to_tensor(self.epsilon, var_dtype),
                neg_lr_t=-apply_state[(var_device, var_dtype)]["lr_t"],
                zero=tf.zeros((), dtype=tf.int64),
            )
        )

    def set_weights(self, weights):
        params = self.weights
        # Override set_weights for backward compatibility of Keras V1 optimizer
        # since it does not include iteration at head of the weight list. Set
        # iteration to 0.
        if len(params) == len(weights) + 1:
            weights = [np.array(0)] + weights
        super().set_weights(weights)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        """Creates an optimizer from its config.

        This method is the reverse of `get_config`,
        capable of instantiating the same optimizer from the config
        dictionary.

        Args:
            config: A Python dictionary, typically the output of get_config.
            custom_objects: A Python dictionary mapping names to additional
              Python objects used to create this optimizer, such as a function
              used for a hyperparameter.

        Returns:
            An optimizer instance.
        """
        if "initial_accumulator_value" not in config:
            config["initial_accumulator_value"] = 0.1
        if "lr" in config:
            config["learning_rate"] = config.pop("lr")
        return cls(**config)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        acc = self.get_slot(var, "accumulator")
        return tf.raw_ops.ResourceApplyAdagradV2(
            var=var.handle,
            accum=acc.handle,
            lr=coefficients["lr_t"],
            epsilon=coefficients["epsilon"],
            grad=grad,
            use_locking=self._use_locking,
        )

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        acc = self.get_slot(var, "accumulator")
        return tf.raw_ops.ResourceSparseApplyAdagradV2(
            var=var.handle,
            accum=acc.handle,
            lr=coefficients["lr_t"],
            epsilon=coefficients["epsilon"],
            grad=grad,
            indices=indices,
            use_locking=self._use_locking,
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "learning_rate": self._serialize_hyperparameter(
                    "learning_rate"
                ),
                "decay": self._initial_decay,
                "initial_accumulator_value": self._initial_accumulator_value,
                "epsilon": self.epsilon,
            }
        )
        return config



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Correctness tests for tf.keras DNN model using DistributionStrategy."""

import numpy as np
import tensorflow.compat.v2 as tf

import keras
from keras import backend
from keras.distribute import keras_correctness_test_base
from keras.distribute import strategy_combinations
from keras.optimizers.optimizer_v2 import (
    gradient_descent as gradient_descent_keras,
)
from keras.testing_infra import test_utils


def all_strategy_combinations_with_eager_and_graph_modes():
    return tf.__internal__.test.combinations.combine(
        distribution=strategy_combinations.all_strategies,
        mode=["graph", "eager"],
    ) + tf.__internal__.test.combinations.combine(
        distribution=strategy_combinations.multi_worker_mirrored_strategies,
        mode="eager",
    )


def all_strategy_combinations_with_graph_mode():
    return tf.__internal__.test.combinations.combine(
        distribution=keras_correctness_test_base.all_strategies, mode=["graph"]
    )


def is_default_strategy(strategy):
    with strategy.scope():
        return not tf.distribute.has_strategy()


@test_utils.run_all_without_tensor_float_32(
    "Uses Dense layers, which call matmul"
)
class TestDistributionStrategyDnnCorrectness(
    keras_correctness_test_base.TestDistributionStrategyCorrectnessBase
):
    def get_model(
        self, initial_weights=None, distribution=None, input_shapes=None
    ):
        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            # We add few non-linear layers to make it non-trivial.
            model = keras.Sequential()
            model.add(
                keras.layers.Dense(10, activation="relu", input_shape=(1,))
            )
            model.add(
                keras.layers.Dense(
                    10,
                    activation="relu",
                    kernel_regularizer=keras.regularizers.l2(1e-4),
                )
            )
            model.add(keras.layers.Dense(10, activation="relu"))
            model.add(keras.layers.Dense(1))

            if initial_weights:
                model.set_weights(initial_weights)

            model.compile(
                loss=keras.losses.mean_squared_error,
                optimizer=gradient_descent_keras.SGD(0.05),
                metrics=["mse"],
            )
            return model

    def get_data(self):
        x_train = np.random.rand(9984, 1).astype("float32")
        y_train = 3 * x_train
        x_predict = np.array([[1.0], [2.0], [3.0], [4.0]], dtype=np.float32)
        return x_train, y_train, x_predict

    def get_data_with_partial_last_batch(self):
        x_train = np.random.rand(10000, 1).astype("float32")
        y_train = 3 * x_train
        x_eval = np.random.rand(10000, 1).astype("float32")
        y_eval = 3 * x_eval
        x_predict = np.array([[1.0], [2.0], [3.0], [4.0]], dtype=np.float32)
        return x_train, y_train, x_eval, y_eval, x_predict

    def get_data_with_partial_last_batch_eval(self):
        x_train = np.random.rand(9984, 1).astype("float32")
        y_train = 3 * x_train
        x_eval = np.random.rand(10000, 1).astype("float32")
        y_eval = 3 * x_eval
        x_predict = np.array([[1.0], [2.0], [3.0], [4.0]], dtype=np.float32)
        return x_train, y_train, x_eval, y_eval, x_predict

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_dnn_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(distribution, use_numpy, use_validation_data)

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_with_tpu_strategies_graph()  # noqa: E501
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_dnn_correctness_with_partial_last_batch_eval(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(
            distribution,
            use_numpy,
            use_validation_data,
            partial_last_batch="eval",
        )

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.strategy_minus_tpu_and_input_config_combinations_eager()  # noqa: E501
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_dnn_correctness_with_partial_last_batch(
        self, distribution, use_numpy, use_validation_data
    ):
        distribution.extended.experimental_enable_get_next_as_optional = True
        self.run_correctness_test(
            distribution,
            use_numpy,
            use_validation_data,
            partial_last_batch="train_and_eval",
            training_epochs=1,
        )

    @tf.__internal__.distribute.combinations.generate(
        all_strategy_combinations_with_graph_mode()
    )
    def test_dnn_with_dynamic_learning_rate(self, distribution):
        self.run_dynamic_lr_test(distribution)


class TestDistributionStrategyDnnMetricCorrectness(
    keras_correctness_test_base.TestDistributionStrategyCorrectnessBase
):
    def get_model(self, distribution=None, input_shapes=None):
        with distribution.scope():
            model = keras.Sequential()
            model.add(
                keras.layers.Dense(
                    1, input_shape=(1,), kernel_initializer="ones"
                )
            )
            model.compile(
                loss=keras.losses.mean_squared_error,
                optimizer=gradient_descent_keras.SGD(0.05),
                metrics=[keras.metrics.BinaryAccuracy()],
            )
        return model

    def run_metric_correctness_test(self, distribution):
        with self.cached_session():
            self.set_up_test_config()

            x_train, y_train, _ = self.get_data()
            model = self.get_model(distribution=distribution)

            batch_size = 64
            batch_size = keras_correctness_test_base.get_batch_size(
                batch_size, distribution
            )
            train_dataset = tf.data.Dataset.from_tensor_slices(
                (x_train, y_train)
            )
            train_dataset = keras_correctness_test_base.batch_wrapper(
                train_dataset, batch_size
            )

            history = model.fit(x=train_dataset, epochs=2, steps_per_epoch=10)
            self.assertEqual(history.history["binary_accuracy"], [1.0, 1.0])

    @tf.__internal__.distribute.combinations.generate(
        all_strategy_combinations_with_eager_and_graph_modes()
    )
    def test_simple_dnn_metric_correctness(self, distribution):
        self.run_metric_correctness_test(distribution)


class TestDistributionStrategyDnnMetricEvalCorrectness(
    keras_correctness_test_base.TestDistributionStrategyCorrectnessBase
):
    def get_model(self, distribution=None, input_shapes=None):
        with distribution.scope():
            model = keras.Sequential()
            model.add(
                keras.layers.Dense(
                    3, activation="relu", input_dim=4, kernel_initializer="ones"
                )
            )
            model.add(
                keras.layers.Dense(
                    1, activation="sigmoid", kernel_initializer="ones"
                )
            )
            model.compile(
                loss="mae",
                metrics=["accuracy", keras.metrics.BinaryAccuracy()],
                optimizer=tf.compat.v1.train.GradientDescentOptimizer(0.001),
            )
        return model

    def run_eval_metrics_correctness_test(self, distribution):
        with self.cached_session():
            self.set_up_test_config()

            model = self.get_model(distribution=distribution)

            # verify correctness of stateful and stateless metrics.
            x = np.ones((100, 4)).astype("float32")
            y = np.ones((100, 1)).astype("float32")
            dataset = tf.data.Dataset.from_tensor_slices((x, y)).repeat()
            dataset = keras_correctness_test_base.batch_wrapper(dataset, 4)
            outs = model.evaluate(dataset, steps=10)
            self.assertEqual(outs[1], 1.0)
            self.assertEqual(outs[2], 1.0)

            y = np.zeros((100, 1)).astype("float32")
            dataset = tf.data.Dataset.from_tensor_slices((x, y)).repeat()
            dataset = keras_correctness_test_base.batch_wrapper(dataset, 4)
            outs = model.evaluate(dataset, steps=10)
            self.assertEqual(outs[1], 0.0)
            self.assertEqual(outs[2], 0.0)

    @tf.__internal__.distribute.combinations.generate(
        all_strategy_combinations_with_eager_and_graph_modes()
    )
    def test_identity_model_metric_eval_correctness(self, distribution):
        self.run_eval_metrics_correctness_test(distribution)


class SubclassedModel(keras.Model):
    def __init__(self, initial_weights, input_shapes):
        super().__init__()
        self.dense1 = keras.layers.Dense(
            10, activation="relu", input_shape=(1,)
        )
        self.dense2 = keras.layers.Dense(
            10,
            activation="relu",
            kernel_regularizer=keras.regularizers.l2(1e-4),
        )
        self.dense3 = keras.layers.Dense(10, activation="relu")
        self.dense4 = keras.layers.Dense(1)
        if input_shapes:
            self.build(input_shapes)
        else:
            # This covers cases when the input is DatasetV1Adapter.
            self.build((None, 1))
        if initial_weights:
            self.set_weights(initial_weights)

    def call(self, inputs):
        x = self.dense1(inputs)
        x = self.dense2(x)
        x = self.dense3(x)
        return self.dense4(x)


@test_utils.run_all_without_tensor_float_32(
    "Uses Dense layers, which call matmul"
)
class TestDistributionStrategyDnnCorrectnessWithSubclassedModel(
    TestDistributionStrategyDnnCorrectness
):
    def get_model(
        self, initial_weights=None, distribution=None, input_shapes=None
    ):
        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            model = SubclassedModel(initial_weights, input_shapes)

            model.compile(
                loss=keras.losses.mean_squared_error,
                optimizer=gradient_descent_keras.SGD(0.05),
                metrics=["mse"],
            )
            return model

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_dnn_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        if (tf.executing_eagerly()) or is_default_strategy(distribution):
            self.run_correctness_test(
                distribution, use_numpy, use_validation_data
            )
        elif (
            backend.is_tpu_strategy(distribution) and not tf.executing_eagerly()
        ):
            with self.assertRaisesRegex(
                ValueError,
                "Expected `model` argument to be a functional `Model` "
                "instance, but got a subclassed model instead.",
            ):
                self.run_correctness_test(
                    distribution, use_numpy, use_validation_data
                )
        else:
            with self.assertRaisesRegex(
                ValueError,
                "We currently do not support distribution strategy with a "
                "`Sequential` model that is created without `input_shape`/"
                "`input_dim` set in its first layer or a subclassed model.",
            ):
                self.run_correctness_test(
                    distribution, use_numpy, use_validation_data
                )

    @tf.__internal__.distribute.combinations.generate(
        all_strategy_combinations_with_graph_mode()
    )
    def test_dnn_with_dynamic_learning_rate(self, distribution):
        if (
            tf.executing_eagerly() and not backend.is_tpu_strategy(distribution)
        ) or is_default_strategy(distribution):
            self.run_dynamic_lr_test(distribution)
        elif backend.is_tpu_strategy(distribution):
            with self.assertRaisesRegex(
                ValueError,
                "Expected `model` argument to be a functional `Model` "
                "instance, but got a subclassed model instead.",
            ):
                self.run_dynamic_lr_test(distribution)
        else:
            with self.assertRaisesRegex(
                ValueError,
                "We currently do not support distribution strategy with a "
                "`Sequential` model that is created without `input_shape`/"
                "`input_dim` set in its first layer or a subclassed model.",
            ):
                self.run_dynamic_lr_test(distribution)

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_with_tpu_strategies_graph()  # noqa: E501
    )
    def test_dnn_correctness_with_partial_last_batch_eval(
        self, distribution, use_numpy, use_validation_data
    ):
        with self.assertRaisesRegex(
            ValueError,
            "Expected `model` argument to be a functional `Model` instance, "
            "but got a subclassed model instead.",
        ):
            self.run_correctness_test(
                distribution,
                use_numpy,
                use_validation_data,
                partial_last_batch="eval",
            )


if __name__ == "__main__":
    tf.__internal__.distribute.multi_process_runner.test_main()



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utils for preprocessing layers."""

import numpy as np
import tensorflow.compat.v2 as tf

from keras.utils import tf_utils

INT = "int"
ONE_HOT = "one_hot"
MULTI_HOT = "multi_hot"
COUNT = "count"
TF_IDF = "tf_idf"


def ensure_tensor(inputs, dtype=None):
    """Ensures the input is a Tensor, SparseTensor or RaggedTensor."""
    if not isinstance(inputs, (tf.Tensor, tf.RaggedTensor, tf.SparseTensor)):
        inputs = tf.convert_to_tensor(inputs, dtype)
    if dtype is not None and inputs.dtype != dtype:
        inputs = tf.cast(inputs, dtype)
    return inputs


def listify_tensors(x):
    """Convert any tensors or numpy arrays to lists for config serialization."""
    if tf.is_tensor(x):
        x = x.numpy()
    if isinstance(x, np.ndarray):
        x = x.tolist()
    return x


def sparse_bincount(inputs, depth, binary_output, dtype, count_weights=None):
    """Apply binary or count encoding to an input and return a sparse tensor."""
    result = tf.sparse.bincount(
        inputs,
        weights=count_weights,
        minlength=depth,
        maxlength=depth,
        axis=-1,
        binary_output=binary_output,
    )
    result = tf.cast(result, dtype)
    if inputs.shape.rank == 1:
        output_shape = (depth,)
    else:
        batch_size = tf.shape(result)[0]
        output_shape = (batch_size, depth)
    result = tf.SparseTensor(
        indices=result.indices, values=result.values, dense_shape=output_shape
    )
    return result


def dense_bincount(inputs, depth, binary_output, dtype, count_weights=None):
    """Apply binary or count encoding to an input."""
    result = tf.math.bincount(
        inputs,
        weights=count_weights,
        minlength=depth,
        maxlength=depth,
        dtype=dtype,
        axis=-1,
        binary_output=binary_output,
    )
    if inputs.shape.rank == 1:
        result.set_shape(tf.TensorShape((depth,)))
    else:
        batch_size = inputs.shape.as_list()[0]
        result.set_shape(tf.TensorShape((batch_size, depth)))
    return result


def expand_dims(inputs, axis):
    """Expand dims on sparse, ragged, or dense tensors."""
    if tf_utils.is_sparse(inputs):
        return tf.sparse.expand_dims(inputs, axis)
    else:
        return tf.expand_dims(inputs, axis)


def encode_categorical_inputs(
    inputs,
    output_mode,
    depth,
    dtype="float32",
    sparse=False,
    count_weights=None,
    idf_weights=None,
):
    """Encodes categoical inputs according to output_mode."""
    if output_mode == INT:
        return tf.identity(tf.cast(inputs, dtype))

    original_shape = inputs.shape
    # In all cases, we should uprank scalar input to a single sample.
    if inputs.shape.rank == 0:
        inputs = expand_dims(inputs, -1)
    # One hot will unprank only if the final output dimension is not already 1.
    if output_mode == ONE_HOT:
        if inputs.shape[-1] != 1:
            inputs = expand_dims(inputs, -1)

    # TODO(b/190445202): remove output rank restriction.
    if inputs.shape.rank > 2:
        raise ValueError(
            f"When output_mode is not `'int'`, maximum supported output rank "
            f"is 2. Received output_mode {output_mode} and input shape "
            f"{original_shape}, "
            f"which would result in output rank {inputs.shape.rank}."
        )

    binary_output = output_mode in (MULTI_HOT, ONE_HOT)
    if sparse:
        bincounts = sparse_bincount(
            inputs, depth, binary_output, dtype, count_weights
        )
    else:
        bincounts = dense_bincount(
            inputs, depth, binary_output, dtype, count_weights
        )

    if output_mode != TF_IDF:
        return bincounts

    if idf_weights is None:
        raise ValueError(
            f"When output mode is `'tf_idf'`, idf_weights must be provided. "
            f"Received: output_mode={output_mode} and idf_weights={idf_weights}"
        )

    if sparse:
        value_weights = tf.gather(idf_weights, bincounts.indices[:, -1])
        return tf.SparseTensor(
            bincounts.indices,
            value_weights * bincounts.values,
            bincounts.dense_shape,
        )
    else:
        return tf.multiply(bincounts, idf_weights)


def compute_shape_for_encode_categorical(shape, output_mode, depth):
    """Computes the output shape of `encode_categorical_inputs`."""
    if output_mode == INT:
        return tf.TensorShape(shape)
    if not shape:
        return tf.TensorShape([depth])
    if output_mode == ONE_HOT and shape[-1] != 1:
        return tf.TensorShape(shape + [depth])
    else:
        return tf.TensorShape(shape[:-1] + [depth])



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Leaky version of a Rectified Linear Unit activation layer."""


from keras import backend
from keras.engine.base_layer import Layer
from keras.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.LeakyReLU")
class LeakyReLU(Layer):
    """Leaky version of a Rectified Linear Unit.

    It allows a small gradient when the unit is not active:

    ```
      f(x) = alpha * x if x < 0
      f(x) = x if x >= 0
    ```

    Usage:

    >>> layer = tf.keras.layers.LeakyReLU()
    >>> output = layer([-3.0, -1.0, 0.0, 2.0])
    >>> list(output.numpy())
    [-0.9, -0.3, 0.0, 2.0]
    >>> layer = tf.keras.layers.LeakyReLU(alpha=0.1)
    >>> output = layer([-3.0, -1.0, 0.0, 2.0])
    >>> list(output.numpy())
    [-0.3, -0.1, 0.0, 2.0]

    Input shape:
      Arbitrary. Use the keyword argument `input_shape`
      (tuple of integers, does not include the batch axis)
      when using this layer as the first layer in a model.

    Output shape:
      Same shape as the input.

    Args:
      alpha: Float >= 0. Negative slope coefficient. Default to 0.3.

    """

    def __init__(self, alpha=0.3, **kwargs):
        super().__init__(**kwargs)
        if alpha is None:
            raise ValueError(
                "The alpha value of a Leaky ReLU layer cannot be None, "
                f"Expecting a float. Received: {alpha}"
            )
        self.supports_masking = True
        self.alpha = backend.cast_to_floatx(alpha)

    def call(self, inputs):
        return backend.relu(inputs, alpha=self.alpha)

    def get_config(self):
        config = {"alpha": float(self.alpha)}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for saving/loading function for keras Model."""

import collections

import keras

# Declaring namedtuple()
ModelFn = collections.namedtuple(
    "ModelFn", ["model", "input_shape", "target_shape"]
)


def basic_sequential():
    """Basic sequential model."""
    model = keras.Sequential(
        [
            keras.layers.Dense(3, activation="relu", input_shape=(3,)),
            keras.layers.Dense(2, activation="softmax"),
        ]
    )
    return ModelFn(model, (None, 3), (None, 2))


def basic_sequential_deferred():
    """Sequential model with deferred input shape."""
    model = keras.Sequential(
        [
            keras.layers.Dense(3, activation="relu"),
            keras.layers.Dense(2, activation="softmax"),
        ]
    )
    return ModelFn(model, (None, 3), (None, 2))


def stacked_rnn():
    """Stacked RNN model."""
    inputs = keras.Input((None, 3))
    layer = keras.layers.RNN([keras.layers.LSTMCell(2) for _ in range(3)])
    x = layer(inputs)
    outputs = keras.layers.Dense(2)(x)
    model = keras.Model(inputs, outputs)
    return ModelFn(model, (None, 4, 3), (None, 2))


def lstm():
    """LSTM model."""
    inputs = keras.Input((None, 3))
    x = keras.layers.LSTM(4, return_sequences=True)(inputs)
    x = keras.layers.LSTM(3, return_sequences=True)(x)
    x = keras.layers.LSTM(2, return_sequences=False)(x)
    outputs = keras.layers.Dense(2)(x)
    model = keras.Model(inputs, outputs)
    return ModelFn(model, (None, 4, 3), (None, 2))


def multi_input_multi_output():
    """Multi-input Multi-output model."""
    body_input = keras.Input(shape=(None,), name="body")
    tags_input = keras.Input(shape=(2,), name="tags")

    x = keras.layers.Embedding(10, 4)(body_input)
    body_features = keras.layers.LSTM(5)(x)
    x = keras.layers.concatenate([body_features, tags_input])

    pred_1 = keras.layers.Dense(2, activation="sigmoid", name="priority")(x)
    pred_2 = keras.layers.Dense(3, activation="softmax", name="department")(x)

    model = keras.Model(
        inputs=[body_input, tags_input], outputs=[pred_1, pred_2]
    )
    return ModelFn(model, [(None, 1), (None, 2)], [(None, 2), (None, 3)])


def nested_sequential_in_functional():
    """A sequential model nested in a functional model."""
    inner_model = keras.Sequential(
        [
            keras.layers.Dense(3, activation="relu", input_shape=(3,)),
            keras.layers.Dense(2, activation="relu"),
        ]
    )

    inputs = keras.Input(shape=(3,))
    x = inner_model(inputs)
    outputs = keras.layers.Dense(2, activation="softmax")(x)
    model = keras.Model(inputs, outputs)
    return ModelFn(model, (None, 3), (None, 2))


def seq_to_seq():
    """Sequence to sequence model."""
    num_encoder_tokens = 3
    num_decoder_tokens = 3
    latent_dim = 2
    encoder_inputs = keras.Input(shape=(None, num_encoder_tokens))
    encoder = keras.layers.LSTM(latent_dim, return_state=True)
    _, state_h, state_c = encoder(encoder_inputs)
    encoder_states = [state_h, state_c]
    decoder_inputs = keras.Input(shape=(None, num_decoder_tokens))
    decoder_lstm = keras.layers.LSTM(
        latent_dim, return_sequences=True, return_state=True
    )
    decoder_outputs, _, _ = decoder_lstm(
        decoder_inputs, initial_state=encoder_states
    )
    decoder_dense = keras.layers.Dense(num_decoder_tokens, activation="softmax")
    decoder_outputs = decoder_dense(decoder_outputs)
    model = keras.Model([encoder_inputs, decoder_inputs], decoder_outputs)
    return ModelFn(
        model,
        [(None, 2, num_encoder_tokens), (None, 2, num_decoder_tokens)],
        (None, 2, num_decoder_tokens),
    )


def shared_layer_functional():
    """Shared layer in a functional model."""
    main_input = keras.Input(shape=(10,), dtype="int32", name="main_input")
    x = keras.layers.Embedding(output_dim=5, input_dim=4, input_length=10)(
        main_input
    )
    lstm_out = keras.layers.LSTM(3)(x)
    auxiliary_output = keras.layers.Dense(
        1, activation="sigmoid", name="aux_output"
    )(lstm_out)
    auxiliary_input = keras.Input(shape=(5,), name="aux_input")
    x = keras.layers.concatenate([lstm_out, auxiliary_input])
    x = keras.layers.Dense(2, activation="relu")(x)
    main_output = keras.layers.Dense(
        1, activation="sigmoid", name="main_output"
    )(x)
    model = keras.Model(
        inputs=[main_input, auxiliary_input],
        outputs=[main_output, auxiliary_output],
    )
    return ModelFn(model, [(None, 10), (None, 5)], [(None, 1), (None, 1)])


def shared_sequential():
    """Shared sequential model in a functional model."""
    inner_model = keras.Sequential(
        [
            keras.layers.Conv2D(2, 3, activation="relu"),
            keras.layers.Conv2D(2, 3, activation="relu"),
        ]
    )
    inputs_1 = keras.Input((5, 5, 3))
    inputs_2 = keras.Input((5, 5, 3))
    x1 = inner_model(inputs_1)
    x2 = inner_model(inputs_2)
    x = keras.layers.concatenate([x1, x2])
    outputs = keras.layers.GlobalAveragePooling2D()(x)
    model = keras.Model([inputs_1, inputs_2], outputs)
    return ModelFn(model, [(None, 5, 5, 3), (None, 5, 5, 3)], (None, 4))


class MySubclassModel(keras.Model):
    """A subclass model."""

    def __init__(self, input_dim=3):
        super().__init__(name="my_subclass_model")
        self._config = {"input_dim": input_dim}
        self.dense1 = keras.layers.Dense(8, activation="relu")
        self.dense2 = keras.layers.Dense(2, activation="softmax")
        self.bn = keras.layers.BatchNormalization()
        self.dp = keras.layers.Dropout(0.5)

    def call(self, inputs, **kwargs):
        x = self.dense1(inputs)
        x = self.dp(x)
        x = self.bn(x)
        return self.dense2(x)

    def get_config(self):
        return self._config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def nested_subclassed_model():
    """A subclass model nested in another subclass model."""

    class NestedSubclassModel(keras.Model):
        """A nested subclass model."""

        def __init__(self):
            super().__init__()
            self.dense1 = keras.layers.Dense(4, activation="relu")
            self.dense2 = keras.layers.Dense(2, activation="relu")
            self.bn = keras.layers.BatchNormalization()
            self.inner_subclass_model = MySubclassModel()

        def call(self, inputs):
            x = self.dense1(inputs)
            x = self.bn(x)
            x = self.inner_subclass_model(x)
            return self.dense2(x)

    return ModelFn(NestedSubclassModel(), (None, 3), (None, 2))


def nested_subclassed_in_functional_model():
    """A subclass model nested in a functional model."""
    inner_subclass_model = MySubclassModel()
    inputs = keras.Input(shape=(3,))
    x = inner_subclass_model(inputs)
    x = keras.layers.BatchNormalization()(x)
    outputs = keras.layers.Dense(2, activation="softmax")(x)
    model = keras.Model(inputs, outputs)
    return ModelFn(model, (None, 3), (None, 2))


def nested_functional_in_subclassed_model():
    """A functional model nested in a subclass model."""

    def get_functional_model():
        inputs = keras.Input(shape=(4,))
        x = keras.layers.Dense(4, activation="relu")(inputs)
        x = keras.layers.BatchNormalization()(x)
        outputs = keras.layers.Dense(2)(x)
        return keras.Model(inputs, outputs)

    class NestedFunctionalInSubclassModel(keras.Model):
        """A functional nested in subclass model."""

        def __init__(self):
            super().__init__(name="nested_functional_in_subclassed_model")
            self.dense1 = keras.layers.Dense(4, activation="relu")
            self.dense2 = keras.layers.Dense(2, activation="relu")
            self.inner_functional_model = get_functional_model()

        def call(self, inputs):
            x = self.dense1(inputs)
            x = self.inner_functional_model(x)
            return self.dense2(x)

    return ModelFn(NestedFunctionalInSubclassModel(), (None, 3), (None, 2))


def shared_layer_subclassed_model():
    """Shared layer in a subclass model."""

    class SharedLayerSubclassModel(keras.Model):
        """A subclass model with shared layers."""

        def __init__(self):
            super().__init__(name="shared_layer_subclass_model")
            self.dense = keras.layers.Dense(3, activation="relu")
            self.dp = keras.layers.Dropout(0.5)
            self.bn = keras.layers.BatchNormalization()

        def call(self, inputs):
            x = self.dense(inputs)
            x = self.dp(x)
            x = self.bn(x)
            return self.dense(x)

    return ModelFn(SharedLayerSubclassModel(), (None, 3), (None, 3))


def functional_with_keyword_args():
    """A functional model with keyword args."""
    inputs = keras.Input(shape=(3,))
    x = keras.layers.Dense(4)(inputs)
    x = keras.layers.BatchNormalization()(x)
    outputs = keras.layers.Dense(2)(x)

    model = keras.Model(inputs, outputs, name="m", trainable=False)
    return ModelFn(model, (None, 3), (None, 2))


ALL_MODELS = [
    ("basic_sequential", basic_sequential),
    ("basic_sequential_deferred", basic_sequential_deferred),
    ("stacked_rnn", stacked_rnn),
    ("lstm", lstm),
    ("multi_input_multi_output", multi_input_multi_output),
    ("nested_sequential_in_functional", nested_sequential_in_functional),
    ("seq_to_seq", seq_to_seq),
    ("shared_layer_functional", shared_layer_functional),
    ("shared_sequential", shared_sequential),
    ("nested_subclassed_model", nested_subclassed_model),
    (
        "nested_subclassed_in_functional_model",
        nested_subclassed_in_functional_model,
    ),
    (
        "nested_functional_in_subclassed_model",
        nested_functional_in_subclassed_model,
    ),
    ("shared_layer_subclassed_model", shared_layer_subclassed_model),
    ("functional_with_keyword_args", functional_with_keyword_args),
]


def get_models(exclude_models=None):
    """Get all models excluding the specified ones."""
    models = [model for model in ALL_MODELS if model[0] not in exclude_models]
    return models



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Adamax optimizer implementation."""

import tensorflow.compat.v2 as tf

from keras import backend_config
from keras.optimizers.optimizer_v2 import optimizer_v2

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.optimizers.Adamax")
class Adamax(optimizer_v2.OptimizerV2):
    """Optimizer that implements the Adamax algorithm.

    It is a variant of Adam based on the infinity norm.
    Default parameters follow those provided in the paper.
    Adamax is sometimes superior to adam, specially in models with embeddings.

    Initialization:

    ```python
    m = 0  # Initialize initial 1st moment vector
    v = 0  # Initialize the exponentially weighted infinity norm
    t = 0  # Initialize timestep
    ```

    The update rule for parameter `w` with gradient `g` is
    described at the end of section 7.1 of the paper:

    ```python
    t += 1
    m = beta1 * m + (1 - beta) * g
    v = max(beta2 * v, abs(g))
    current_lr = learning_rate / (1 - beta1 ** t)
    w = w - current_lr * m / (v + epsilon)
    ```

    Similarly to `Adam`, the epsilon is added for numerical stability
    (especially to get rid of division by zero when `v_t == 0`).

    In contrast to `Adam`, the sparse implementation of this algorithm
    (used when the gradient is an IndexedSlices object, typically because of
    `tf.gather` or an embedding lookup in the forward pass) only updates
    variable slices and corresponding `m_t`, `v_t` terms when that part of
    the variable was used in the forward pass. This means that the sparse
    behavior is contrast to the dense behavior (similar to some momentum
    implementations which ignore momentum unless a variable slice was actually
    used).

    Args:
      learning_rate: A `Tensor`, floating point value, or a schedule that is a
        `tf.keras.optimizers.schedules.LearningRateSchedule`. The learning rate.
      beta_1: A float value or a constant float tensor. The exponential decay
        rate for the 1st moment estimates.
      beta_2: A float value or a constant float tensor. The exponential decay
        rate for the exponentially weighted infinity norm.
      epsilon: A small constant for numerical stability.
      name: Optional name for the operations created when applying gradients.
        Defaults to `"Adamax"`.
      **kwargs: keyword arguments. Allowed arguments are `clipvalue`,
        `clipnorm`, `global_clipnorm`.
        If `clipvalue` (float) is set, the gradient of each weight
        is clipped to be no higher than this value.
        If `clipnorm` (float) is set, the gradient of each weight
        is individually clipped so that its norm is no higher than this value.
        If `global_clipnorm` (float) is set the gradient of all weights is
        clipped so that their global norm is no higher than this value.

    Reference:
      - [Kingma et al., 2014](http://arxiv.org/abs/1412.6980)
    """

    _HAS_AGGREGATE_GRAD = True

    def __init__(
        self,
        learning_rate=0.001,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=1e-7,
        name="Adamax",
        **kwargs
    ):
        super().__init__(name, **kwargs)
        self._set_hyper("learning_rate", kwargs.get("lr", learning_rate))
        self._set_hyper("decay", self._initial_decay)
        self._set_hyper("beta_1", beta_1)
        self._set_hyper("beta_2", beta_2)
        self.epsilon = epsilon or backend_config.epsilon()

    def _create_slots(self, var_list):
        # Separate for-loops to respect the ordering of slot variables from v1.
        for var in var_list:
            self.add_slot(var, "m")  # Create slots for the first moments.
        for var in var_list:
            self.add_slot(var, "v")  # Create slots for the second moments.

    def _prepare_local(self, var_device, var_dtype, apply_state):
        super()._prepare_local(var_device, var_dtype, apply_state)

        local_step = tf.cast(self.iterations + 1, var_dtype)
        beta_1_t = tf.identity(self._get_hyper("beta_1", var_dtype))
        beta_2_t = tf.identity(self._get_hyper("beta_2", var_dtype))
        beta_1_power = tf.pow(beta_1_t, local_step)
        lr_t = apply_state[(var_device, var_dtype)]["lr_t"]

        apply_state[(var_device, var_dtype)].update(
            dict(
                neg_scaled_lr=-lr_t / (1 - beta_1_power),
                epsilon=tf.convert_to_tensor(self.epsilon, var_dtype),
                beta_1_t=beta_1_t,
                beta_1_power=beta_1_power,
                one_minus_beta_1_t=1 - beta_1_t,
                beta_2_t=beta_2_t,
                zero=tf.zeros((), dtype=tf.int64),
            )
        )

    def _resource_apply_dense(self, grad, var, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        m = self.get_slot(var, "m")
        v = self.get_slot(var, "v")
        return tf.raw_ops.ResourceApplyAdaMax(
            var=var.handle,
            m=m.handle,
            v=v.handle,
            beta1_power=coefficients["beta_1_power"],
            lr=coefficients["lr_t"],
            beta1=coefficients["beta_1_t"],
            beta2=coefficients["beta_2_t"],
            epsilon=coefficients["epsilon"],
            grad=grad,
            use_locking=self._use_locking,
        )

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        var_device, var_dtype = var.device, var.dtype.base_dtype
        coefficients = (apply_state or {}).get(
            (var_device, var_dtype)
        ) or self._fallback_apply_state(var_device, var_dtype)

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_slice = tf.gather(m, indices, axis=coefficients["zero"])
        m_t_slice = (
            m_slice * coefficients["beta_1_t"]
            + grad * coefficients["one_minus_beta_1_t"]
        )
        with tf.control_dependencies([m_t_slice]):
            m_t = self._resource_scatter_update(m, indices, m_t_slice)

        # u_t = max(beta2 * u, abs(g_t))
        v = self.get_slot(var, "v")
        v_slice = tf.gather(v, indices, axis=coefficients["zero"])
        v_t_slice = tf.maximum(v_slice * coefficients["beta_2_t"], tf.abs(grad))
        with tf.control_dependencies([v_t_slice]):
            v_t = self._resource_scatter_update(v, indices, v_t_slice)
        # theta_t = theta - lr / (1 - beta1^t) * m_t / u_t
        var_slice = coefficients["neg_scaled_lr"] * (
            m_t_slice / (v_t_slice + coefficients["epsilon"])
        )
        with tf.control_dependencies([var_slice]):
            var_update = self._resource_scatter_add(var, indices, var_slice)
        return tf.group(*[var_update, m_t, v_t])

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "learning_rate": self._serialize_hyperparameter(
                    "learning_rate"
                ),
                "decay": self._initial_decay,
                "beta_1": self._serialize_hyperparameter("beta_1"),
                "beta_2": self._serialize_hyperparameter("beta_2"),
                "epsilon": self.epsilon,
            }
        )
        return config



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Contains the InputSpec class."""

import tensorflow.compat.v2 as tf

from keras import backend

# isort: off
from tensorflow.python.util.tf_export import keras_export
from tensorflow.python.util.tf_export import tf_export


@keras_export(
    "keras.layers.InputSpec",
    v1=["keras.layers.InputSpec", "keras.__internal__.legacy.layers.InputSpec"],
)
@tf_export(v1=["layers.InputSpec"])
class InputSpec:
    """Specifies the rank, dtype and shape of every input to a layer.

    Layers can expose (if appropriate) an `input_spec` attribute:
    an instance of `InputSpec`, or a nested structure of `InputSpec` instances
    (one per input tensor). These objects enable the layer to run input
    compatibility checks for input structure, input rank, input shape, and
    input dtype.

    A None entry in a shape is compatible with any dimension,
    a None shape is compatible with any shape.

    Args:
      dtype: Expected DataType of the input.
      shape: Shape tuple, expected shape of the input
        (may include None for unchecked axes). Includes the batch size.
      ndim: Integer, expected rank of the input.
      max_ndim: Integer, maximum rank of the input.
      min_ndim: Integer, minimum rank of the input.
      axes: Dictionary mapping integer axes to
        a specific dimension value.
      allow_last_axis_squeeze: If True, then allow inputs of rank N+1 as long
        as the last axis of the input is 1, as well as inputs of rank N-1
        as long as the last axis of the spec is 1.
      name: Expected key corresponding to this input when passing data as
        a dictionary.

    Example:

    ```python
    class MyLayer(Layer):
        def __init__(self):
            super(MyLayer, self).__init__()
            # The layer will accept inputs with
            # shape (?, 28, 28) & (?, 28, 28, 1)
            # and raise an appropriate error message otherwise.
            self.input_spec = InputSpec(
                shape=(None, 28, 28, 1),
                allow_last_axis_squeeze=True)
    ```
    """

    def __init__(
        self,
        dtype=None,
        shape=None,
        ndim=None,
        max_ndim=None,
        min_ndim=None,
        axes=None,
        allow_last_axis_squeeze=False,
        name=None,
    ):
        self.dtype = tf.as_dtype(dtype).name if dtype is not None else None
        shape = tf.TensorShape(shape)
        if shape.rank is None:
            shape = None
        else:
            shape = tuple(shape.as_list())
        if shape is not None:
            self.ndim = len(shape)
            self.shape = shape
        else:
            self.ndim = ndim
            self.shape = None
        self.max_ndim = max_ndim
        self.min_ndim = min_ndim
        self.name = name
        self.allow_last_axis_squeeze = allow_last_axis_squeeze
        try:
            axes = axes or {}
            self.axes = {int(k): axes[k] for k in axes}
        except (ValueError, TypeError):
            raise TypeError(
                "Argument `axes` must be a dict with integer keys. "
                f"Received: axes={axes}"
            )

        if self.axes and (self.ndim is not None or self.max_ndim is not None):
            max_dim = (self.ndim if self.ndim else self.max_ndim) - 1
            max_axis = max(self.axes)
            if max_axis > max_dim:
                raise ValueError(
                    "Axis {} is greater than the maximum "
                    "allowed value: {}".format(max_axis, max_dim)
                )

    def __repr__(self):
        spec = [
            ("dtype=" + str(self.dtype)) if self.dtype else "",
            ("shape=" + str(self.shape)) if self.shape else "",
            ("ndim=" + str(self.ndim)) if self.ndim else "",
            ("max_ndim=" + str(self.max_ndim)) if self.max_ndim else "",
            ("min_ndim=" + str(self.min_ndim)) if self.min_ndim else "",
            ("axes=" + str(self.axes)) if self.axes else "",
        ]
        return "InputSpec(%s)" % ", ".join(x for x in spec if x)

    def get_config(self):
        return {
            "dtype": self.dtype,
            "shape": self.shape,
            "ndim": self.ndim,
            "max_ndim": self.max_ndim,
            "min_ndim": self.min_ndim,
            "axes": self.axes,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def to_tensor_shape(spec):
    """Returns a tf.TensorShape object that matches the shape specifications.

    If the InputSpec's shape or ndim is defined, this method will return a fully
    or partially-known shape. Otherwise, the returned TensorShape is None.

    Args:
      spec: an InputSpec object.

    Returns:
      a tf.TensorShape object
    """
    if spec.ndim is None and spec.shape is None:
        return tf.TensorShape(None)
    elif spec.shape is not None:
        return tf.TensorShape(spec.shape)
    else:
        shape = [None] * spec.ndim
        for a in spec.axes:
            shape[a] = spec.axes[a]  # Assume that axes is defined
        return tf.TensorShape(shape)


def assert_input_compatibility(input_spec, inputs, layer_name):
    """Checks compatibility between the layer and provided inputs.

    This checks that the tensor(s) `inputs` verify the input assumptions
    of a layer (if any). If not, a clear and actional exception gets raised.

    Args:
        input_spec: An InputSpec instance, list of InputSpec instances, a nested
            structure of InputSpec instances, or None.
        inputs: Input tensor, list of input tensors, or a nested structure of
            input tensors.
        layer_name: String, name of the layer (for error message formatting).

    Raises:
        ValueError: in case of mismatch between
            the provided inputs and the expectations of the layer.
    """
    if not input_spec:
        return

    input_spec = tf.nest.flatten(input_spec)
    if isinstance(inputs, dict):
        # Flatten `inputs` by reference order if input spec names are provided
        names = [spec.name for spec in input_spec]
        if all(names):
            list_inputs = []
            for name in names:
                if name not in inputs:
                    raise ValueError(
                        f'Missing data for input "{name}". '
                        "You passed a data dictionary with keys "
                        f"{list(inputs.keys())}. "
                        f"Expected the following keys: {names}"
                    )
                list_inputs.append(inputs[name])
            inputs = list_inputs

    inputs = tf.nest.flatten(inputs)
    for x in inputs:
        # Having a shape/dtype is the only commonality of the various
        # tensor-like objects that may be passed. The most common kind of
        # invalid type we are guarding for is a Layer instance (Functional API),
        # which does not have a `shape` attribute.
        if not hasattr(x, "shape"):
            raise TypeError(f"Inputs to a layer should be tensors. Got: {x}")

    if len(inputs) != len(input_spec):
        raise ValueError(
            f'Layer "{layer_name}" expects {len(input_spec)} input(s),'
            f" but it received {len(inputs)} input tensors. "
            f"Inputs received: {inputs}"
        )
    for input_index, (x, spec) in enumerate(zip(inputs, input_spec)):
        if spec is None:
            continue

        shape = tf.TensorShape(x.shape)
        if shape.rank is None:
            return
        # Check ndim.
        if spec.ndim is not None and not spec.allow_last_axis_squeeze:
            ndim = shape.rank
            if ndim != spec.ndim:
                raise ValueError(
                    f'Input {input_index} of layer "{layer_name}" '
                    "is incompatible with the layer: "
                    f"expected ndim={spec.ndim}, found ndim={ndim}. "
                    f"Full shape received: {tuple(shape)}"
                )
        if spec.max_ndim is not None:
            ndim = x.shape.rank
            if ndim is not None and ndim > spec.max_ndim:
                raise ValueError(
                    f'Input {input_index} of layer "{layer_name}" '
                    "is incompatible with the layer: "
                    f"expected max_ndim={spec.max_ndim}, "
                    f"found ndim={ndim}"
                )
        if spec.min_ndim is not None:
            ndim = x.shape.rank
            if ndim is not None and ndim < spec.min_ndim:
                raise ValueError(
                    f'Input {input_index} of layer "{layer_name}" '
                    "is incompatible with the layer: "
                    f"expected min_ndim={spec.min_ndim}, "
                    f"found ndim={ndim}. "
                    f"Full shape received: {tuple(shape)}"
                )
        # Check dtype.
        if spec.dtype is not None:
            if x.dtype.name != spec.dtype:
                raise ValueError(
                    f'Input {input_index} of layer "{layer_name}" '
                    "is incompatible with the layer: "
                    f"expected dtype={spec.dtype}, "
                    f"found dtype={x.dtype}"
                )

        # Check specific shape axes.
        shape_as_list = shape.as_list()
        if spec.axes:
            for axis, value in spec.axes.items():
                if hasattr(value, "value"):
                    value = value.value
                if value is not None and shape_as_list[int(axis)] not in {
                    value,
                    None,
                }:
                    raise ValueError(
                        f'Input {input_index} of layer "{layer_name}" is '
                        f"incompatible with the layer: expected axis {axis} "
                        f"of input shape to have value {value}, "
                        "but received input with "
                        f"shape {display_shape(x.shape)}"
                    )
        # Check shape.
        if spec.shape is not None and shape.rank is not None:
            spec_shape = spec.shape
            if spec.allow_last_axis_squeeze:
                if shape_as_list and shape_as_list[-1] == 1:
                    shape_as_list = shape_as_list[:-1]
                if spec_shape and spec_shape[-1] == 1:
                    spec_shape = spec_shape[:-1]
            for spec_dim, dim in zip(spec_shape, shape_as_list):
                if spec_dim is not None and dim is not None:
                    if spec_dim != dim:
                        raise ValueError(
                            f'Input {input_index} of layer "{layer_name}" is '
                            "incompatible with the layer: "
                            f"expected shape={spec.shape}, "
                            f"found shape={display_shape(x.shape)}"
                        )


def display_shape(shape):
    return str(tuple(shape.as_list()))


def to_tensor_spec(input_spec, default_dtype=None):
    """Converts a Keras InputSpec object to a TensorSpec."""
    default_dtype = default_dtype or backend.floatx()
    if isinstance(input_spec, InputSpec):
        dtype = input_spec.dtype or default_dtype
        return tf.TensorSpec(to_tensor_shape(input_spec), dtype)
    return tf.TensorSpec(None, default_dtype)
